/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2007-2018 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: target_config.c 1326 2019-01-26 15:52:57Z ertl-honda $  
 */

/*
 *  ターゲット依存モジュール（i.MX8 Cortex-A53(AArch64)用）
 */
#include "kernel_impl.h"
#include <sil.h>
#include "imx8_uart.h"

#ifdef TOPPERS_WITH_ATF
#include <atf.h>
#include <psci.h>
#endif /* TOPPERS_WITH_ATF */

/*
 *  entry point (start.S)
 */
extern void start(void);

/* ATFのTrusted OSベクタテーブル */
extern atfsmc_vectors_t atf_vector_table;

/*
 *  EL3で行う初期化処理
 */
void
target_mon_initialize(void)
{
	volatile uint32_t   reg32_val;

	SCR_EL3_READ(reg32_val);
	reg32_val &= ~(SCR_EA_BIT|SCR_FIQ_BIT|SCR_IRQ_BIT|SCR_NS_BIT);
	SCR_EL3_WRITE(reg32_val);

	/*
	 *  GIC System register enable
	 */
	ICC_SRE_EL3_READ(reg32_val);
	reg32_val |= (uint32_t)((1 << 3) | (1 << 0));
	ICC_SRE_EL3_WRITE(reg32_val);
}

/*
 *  EL2で行う初期化処理
 */
void
target_hyp_initialize(void)
{
	volatile uint32_t   reg32_val;

	/*
	 *  GIC System register enable
	 */
	ICC_SRE_EL2_READ(reg32_val);
	reg32_val |= (uint32_t)((1 << 3) | (1 << 0));
	ICC_SRE_EL2_WRITE(reg32_val);

	/* CPUECTLR_EL1レジスタのNS-EL1からのアクセス許可 */
	ACTLR_EL2_READ(reg32_val);
	reg32_val |= (1 << 1);
	ACTLR_EL2_WRITE(reg32_val);

	/*
	 *  Generic Timerの初期化
	 */
	/* Physical Counter, Physical TimerをEL1NSとEL0NSからアクセス可能に */
	CNTHCTL_EL2_WRITE(CNTHCTL_EL1PCEN_BIT | CNTHCTL_EL1PCTEN_BIT);

	/* Virtual Counterのオフセットを0に */
	CNTVOFF_EL2_WRITE(0);

	inst_sync_barrier();
}

/*
 *  str_ker() の前でマスタプロセッサで行う初期化
 */
void
target_mprc_initialize(void)
{
	uint32_t i;

	dcache_disable();
	icache_disable();

	chip_mprc_initialize();

#ifdef TOPPERS_WITH_ATF
#ifdef TOPPERS_TZ_S
#error
#else /* TOPPERS_TZ_NS */

	/*
	 *  ARM Trusted Firmware(None Secure) 使用時
	 */

	/*
	 *  サブコア ON
	 */
	for(i = 1; i < TNUM_PRCID; i++) {
		psci_smc_cpuon(i, (uint64_t)start, 0);
	}

#endif /* TOPPERS_TZ_S */
#else  /* !TOPPERS_WITH_ATF */
#error   
#endif /* TOPPERS_WITH_ATF */

#ifdef TOPPERS_BOOT_M4
	/*
	 *  Cortex-M4 On
	 */
	/* Enable ARM_M4_CLK ROOT */
	/* Enable Cortex-M4 */
	reg32_val = sil_rew_mem((void*)IMX8REG_SRC_M4RCR);
	sil_wrw_mem((void*)IMX8REG_SRC_M4RCR, reg32_val | (1 << 3));
	/* Reset Cortex-M4 core and platform */
	reg32_val = sil_rew_mem((void*)IMX8REG_SRC_M4RCR);
	sil_wrw_mem((void*)IMX8REG_SRC_M4RCR, reg32_val | (1 << 2) | (1 << 1));
	do {
		reg32_val = sil_rew_mem((void*)IMX8REG_SRC_M4RCR);
	} while( (reg32_val & ((1 << 2) | (1 << 1))) != 0 );
#endif /* TOPPERS_BOOT_M4 */
}

/*
 *  ユーザープログラムによるMMUの初期化
 */
void user_mmu_init(void) __attribute__((weak));

/*
 *  ユーザーが定義しなかった場合のweak関数
 */
void
user_mmu_init(void) {
}

/*
 *  メモリマップの初期化（core依存部から呼び出される）
 *  同時にキャッシュも有効にする
 */
void
target_mmu_init(void)
{
	mmap_t mm;

	/*
	 *  メモリマップの設定
	 *  全領域は物理アドレス = 仮想アドレス
	 */
	/* Other */
	mm.pa   = 0x000000000000;
	mm.va   = mm.pa;
	mm.size = 0x000000900000;
	mm.attr = MEM_ATTR_DEV;
	mm.ns   = MEM_NS_NONSECURE;
	mmu_mmap_add(&mm);

	/* OCRAM */
	mm.pa   = 0x000000900000;
	mm.va   = mm.pa;
	mm.size = 0x000000020000;
	mm.attr = MEM_ATTR_NML_C;
	mm.ns   = MEM_NS_NONSECURE;
	mmu_mmap_add(&mm);

	/* Other */
	mm.pa   = 0x000008000000;
	mm.va   = mm.pa;
	mm.size = 0x000038000000;
	mm.attr = MEM_ATTR_DEV;
	mm.ns   = MEM_NS_NONSECURE;
	mmu_mmap_add(&mm);

	/* DDR memory */
	mm.pa   = 0x0040000000;
	mm.va   = mm.pa;
	mm.size = 0x0010000000;
	mm.attr = MEM_ATTR_NML_C;
	mm.ns   = MEM_NS_SECURE;
	mmu_mmap_add(&mm);

	/*
	 *  ユーザープログラムによるMMUの初期化
	 */
	user_mmu_init();
}

/*
 *  ターゲット依存の初期化
 */
void
target_initialize(void)
{
	/*
	 * チップ依存の初期化
	 */
	chip_initialize();

	/*
	 *  バナー表示，低レベル出力用にUARTを初期化
	 */
	if (x_sense_mprc()) {
		sio_low_init(1);
	}
}

/*
 *  ターゲット依存の終了処理
 */
void
target_exit(void)
{
	/*
	 * チップ依存の終了処理
	 */
	chip_exit();

	while(1);
}

/*
 *  システムログの低レベル出力のための文字出力
 */
void
target_fput_log(char c)
{
	if (c == '\n') {
		sio_pol_putc('\r', 1);
	}
	sio_pol_putc(c, 1);
}
